/*
 * ActiveLearner.java
 *
 * The bootstrapper using this class must have a relation that equals to
 * 'None of the others', i.e., the relation list must be mutually-exclusive
 * and also exhaustive.
 *
 * Revision History:
 *  4/12/2006  Filled in pretty much everything.
 *   Andy thinks C# is dreamy. Slowly becoming a Java convert.
 *    He's hooked on shiny IDEs... and hungry.
 *   Kevin is in love with Perl. and hungry.
 *   Sue Ann's hungry.
 *   Sophie's cold.
 */

package activelearner;
import java.util.*;

public class ActiveLearner {
    public enum Arity { UNARY, BINARY };
    
    /**
     * Class data members (pretty much all private)
     **/
    public static String DUNNO = "I do not know.";
    static String SEED = "*SEED*";
    // these members keep track of what the bootstrapper told us
    Arity arity;
    List<String> relations;
    List<String> rules;
    List<String> entities;
    // HashMap<rule, HashMap<entity,counts>>
    HashMap<String,HashMap<String, Integer>> rule2EntityCountHash;
    // HashMap<entity, HashMap<rule,counts>>
    HashMap<String,HashMap<String, Integer>> entity2RuleCountHash;
    // Where rules came from.
    HashMap<String,List<String>> ruleFromEntities; // <rule, entity list>
    // What seed rules were meant for.
    HashMap<String,String> ruleForRelations; // <rule, relation>
    // What seed entities were meant for.
    HashMap<String,String> entityForRelations; // <entity, relation>
    
    // these members store the results of our algorithm
    // HashMap<rule, HashMap<relation,score>>
    HashMap<String,HashMap<String, Double>> ruleScores;
    // HashMap<entity, HashMap<relation,score>>
    HashMap<String,HashMap<String, Double>> entityScores;
    List<String> rankedEntities;
    
    // Keeping track of questions asked/answered
    // TODO: Ranked list of entities/questions?
    HashMap<String,Boolean> asked;     // HashMap<entity,yes/no>
    HashMap<String,Boolean> answered;  // HashMap<entity,yes/no>
    HashMap<String,String> answers;    // HashMap<entity,relation/dunno>
    // TODO: We need to keep track of what we've been told by user
    //List<Fact> knownFacts;
    
    
    /**
     * Constructor-ish Things
     **/
    /* Create a new active learner object, specifying the
     * list of mutually exclusive relations that are to be
     * learned. */
    public ActiveLearner(List<String> relations, Arity arity ) {
        this.registerRelations(relations, arity);
        
        this.rules = new ArrayList<String>();
        this.ruleScores = new HashMap<String,HashMap<String,Double>>();
        this.entityScores = new HashMap<String,HashMap<String,Double>>();
        this.entity2RuleCountHash = new HashMap<String,HashMap<String,Integer>>();
        this.rule2EntityCountHash = new HashMap<String,HashMap<String,Integer>>();
        this.entities = new ArrayList<String>();
        
        this.ruleForRelations = new HashMap<String,String>();
        this.ruleFromEntities = new HashMap<String,List<String>>();
        
        this.asked = new HashMap<String,Boolean>();
        this.answered = new HashMap<String,Boolean>();
        this.answers = new HashMap<String,String>();
        
        this.entityForRelations = new HashMap<String,String>();
    
    }
    private void registerRelations( List<String> relations, Arity arity ) {
        assert(relations.size() > 1);
        this.relations = new ArrayList<String>(relations);
        this.arity = arity;
    }
    
    
    /**
     * UI-ish functions; dealing with question and answers
     **/
    /* A question consists of a question string and a list of
     * possible answers.  The getQuestion should be called by
     * the user interface when there is a user waiting for a
     * question.  When the user answers the question, the
     * interface should call answerQuestion with the index
     * of the checkbox checked by the user.
     *
     * Returns null if we have no more questions to ask.
     */
    public Question getQuestion() {
        int blah = rankEntitiesForQuestions();
        String entity = null;
        boolean gotQuestion = false;
        
        for (String ent: rankedEntities) {
            // Ask only if it's not been asked already.
            if (!(asked.containsKey(ent) && asked.get(ent))) {
                entity = ent;
                asked.put(entity, true);     // HashMap<entity,yes/no>
                answered.put(entity,false);  // HashMap<entity,yes/no>
                break;
            }
        }
        
        if (entity == null) return null;
        else {
            List<String> amendedRelations = new ArrayList<String>(relations);
            amendedRelations.add(this.DUNNO);
            return new Question(arity, entity, amendedRelations);
        }
    }
    public int getNumQuestions() {
        return rankEntitiesForQuestions();
    }
    public void answerQuestion( Question ques ) {
        String answer = ques.getAnswer();
        if (answer != null) {
            String entity = ques.getEntity();
            answered.remove(entity);
            answered.put(entity,true);
            answers.put(entity,answer);
        }
    }
    
    /**
     * Adding stuff
     **/
    /* Rules are either specified as seeds for a certain relation
     * or extracted by a set of entities.  The addSeedRule function
     * requires that the client specify which relation the rule is
     * intended to extract.  The addRule function requires that the
     * client specify which entities were used in the extraction of
     * the rule. */
    public void addSeedRule( String rule, String relation ) {
        ruleForRelations.put(rule, relation); // <rule, relation>
        List<String> fromEntities = new ArrayList(1);
        fromEntities.add(SEED);
        addRule(rule, fromEntities);
    }
    // If a rule had been added already, we ignore this add.
    public void addRule( String rule, List<String> fromEntities ) {
        if (rules.contains(rule)) return;
        
        this.rules.add(rule);
        putDefaultScoresInRuleHash(rule);
        // This should only be called once for a rule.
        ruleFromEntities.put(rule,new ArrayList(fromEntities));
        this.rule2EntityCountHash.put(rule, new HashMap<String,Integer>());
    }
    
    /* Seeds are either specified as seeds for a certain relation
     * or extracted by a rule.  The addSeedEntity function requires
     * that the client specify the relation to which the entity
     * belongs.  The addEntityCount requires that the client specify
     * both the rule that extracted the entity and the number of
     * times the entity was extracted by the rule. */
    public void addSeedEntity( String entity, String relation ) {
        if (!entities.contains(entity)) { entities.add(entity); }
        entityForRelations.put(entity, relation); 
        // this hashmap keeps track of seed entities.  it's not used by anything else yet.
        if (!entity2RuleCountHash.containsKey(entity)) {
            this.entity2RuleCountHash.put(entity, new HashMap<String,Integer>());
            putDefaultScoresInEntityHash(entity);   
        }
    }
    public void addEntityCount( String rule, String entity, int count ) {
        if (!entities.contains(entity)) { entities.add(entity); }
        if (!entity2RuleCountHash.containsKey(entity)) {
            this.entity2RuleCountHash.put(entity, new HashMap<String,Integer>());
            putDefaultScoresInEntityHash(entity);
        }
        
        // We must update all hashes
        int currentCount = 0;
        if (this.entity2RuleCountHash.get(entity).containsKey(rule)) {
            currentCount = this.entity2RuleCountHash.get(entity).get(rule);
        }
        currentCount += count;
        this.entity2RuleCountHash.get(entity).put(rule,currentCount);
        this.rule2EntityCountHash.get(rule).put(entity,currentCount);
    }
    
    /* A probability or confidence score is associated with every
     * entity and every rule.  The getEntityScore function returns
     * a number between zero and one.  A high score reflects that
     * the active learner considers it likely that the given relation
     * holds for the entity.  A low score means the active learner
     * considers it unlikely.  Likewise, the getRuleScore function
     * returns a number between zero and one, and a higher number
     * indicates a higher confidence that the rule is a good extractor
     * for the given relation. */
    public double getEntityScore( String entity, String relation ) {
        scoreEstimator(); // Call the function to update scores
        return entityScores.get(entity).get(relation);
    }
    public double getRuleScore( String rule, String relation ) {
        scoreEstimator(); // Call the function to update scores
        return ruleScores.get(rule).get(relation);
    }
    
    
    /**
     * Real functions!!
     **/
    // Return the number of questions we want answered now.
    private int rankEntitiesForQuestions() {
        // TODO
        rankedEntities = new ArrayList<String>(entities);
        return 1;
    }
    // Set values in ruleScores and entityScores to something sensible.
    private void scoreEstimator() {
        // Remember there are entities and rules with empty score hashmaps.
        
    }
    

    private void putDefaultScoresInRuleHash(String rule) {
        HashMap<String,Double> newHash = new HashMap<String, Double>();
        for (String rel: relations) {
            newHash.put(rel,0.5);
        }
        ruleScores.put(rule,newHash);
    }
    private void putDefaultScoresInEntityHash(String entity) {
        HashMap<String,Double> newHash = new HashMap<String, Double>();
        for (String rel: relations) {
            newHash.put(rel,0.5);
        }
        entityScores.put(entity,newHash);
    }
    
}
